package java.awt.font;

import sun.font.FontResolver;
import sun.text.CodePointIterator;
import sun.font.GraphicComponent;
import sun.font.ExtendedTextLabel;
import sun.font.Decoration;
import sun.font.TextLabelFactory;
import java.text.Bidi;
import sun.font.AttributeValues;
import java.text.AttributedCharacterIterator;
import java.util.Map;
import java.awt.Font;
import java.awt.geom.GeneralPath;
import java.awt.Shape;
import sun.font.BidiUtils;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.awt.Color;
import java.awt.image.BufferedImage;
import java.awt.geom.AffineTransform;
import sun.font.CoreMetrics;
import java.awt.geom.Point2D;
import java.awt.Rectangle;
import sun.font.LayoutPathImpl;
import sun.font.TextLineComponent;

final class TextLine
{
    private TextLineComponent[] fComponents;
    private float[] fBaselineOffsets;
    private int[] fComponentVisualOrder;
    private float[] locs;
    private char[] fChars;
    private int fCharsStart;
    private int fCharsLimit;
    private int[] fCharVisualOrder;
    private int[] fCharLogicalOrder;
    private byte[] fCharLevels;
    private boolean fIsDirectionLTR;
    private LayoutPathImpl lp;
    private boolean isSimple;
    private Rectangle pixelBounds;
    private FontRenderContext frc;
    private TextLine.TextLineMetrics fMetrics;
    private static TextLine.Function fgPosAdvF;
    private static TextLine.Function fgAdvanceF;
    private static TextLine.Function fgXPositionF;
    private static TextLine.Function fgYPositionF;
    
    public TextLine(FontRenderContext frc, TextLineComponent[] fComponents, float[] fBaselineOffsets, char[] fChars, int fCharsStart, int fCharsLimit, int[] fCharLogicalOrder, byte[] fCharLevels, boolean fIsDirectionLTR) {
        this.fMetrics = null;
        int[] computeComponentOrder = computeComponentOrder(fComponents, fCharLogicalOrder);
        this.frc = frc;
        this.fComponents = fComponents;
        this.fBaselineOffsets = fBaselineOffsets;
        this.fComponentVisualOrder = computeComponentOrder;
        this.fChars = fChars;
        this.fCharsStart = fCharsStart;
        this.fCharsLimit = fCharsLimit;
        this.fCharLogicalOrder = fCharLogicalOrder;
        this.fCharLevels = fCharLevels;
        this.fIsDirectionLTR = fIsDirectionLTR;
        this.checkCtorArgs();
        this.init();
    }
    
    private void checkCtorArgs() {
        int n = 0;
        for (int i = 0; i < this.fComponents.length; ++i) {
            n += this.fComponents[i].getNumCharacters();
        }
        if (n != this.characterCount()) {
            throw new IllegalArgumentException("Invalid TextLine!  char count is different from sum of char counts of components.");
        }
        return;
    }
    
    private void init() {
        float max = 0.0f;
        float max2 = 0.0f;
        float max3 = 0.0f;
        float max4 = 0.0f;
        float max5 = 0.0f;
        boolean n = false;
        this.isSimple = true;
        for (int i = 0; i < this.fComponents.length; ++i) {
            TextLineComponent textLineComponent = this.fComponents[i];
            this.isSimple &= textLineComponent.isSimple();
            CoreMetrics coreMetrics = textLineComponent.getCoreMetrics();
            int n2 = (byte)coreMetrics.baselineIndex;
            if (n2 >= 0) {
                float n3 = this.fBaselineOffsets[n2];
                max = Math.max(max, -n3 + coreMetrics.ascent);
                float b3 = n3 + coreMetrics.descent;
                max2 = Math.max(max2, b3);
                max3 = Math.max(max3, b3 + coreMetrics.leading);
            }
            else {
                n = true;
                float b = coreMetrics.ascent + coreMetrics.descent;
                float b2 = b + coreMetrics.leading;
                max4 = Math.max(max4, b);
                max5 = Math.max(max5, b2);
            }
        }
        if (!(n == 0)) {
            if (max4 > max + max2) {
                max2 = max4 - max;
            }
            if (max5 > max + max3) {
                max3 = max5 - max;
            }
        }
        float n4 = max3 - max2;
        if (!(n == 0)) {
            this.fBaselineOffsets = new float[] { this.fBaselineOffsets[0], this.fBaselineOffsets[1], this.fBaselineOffsets[2], max2, -max };
        }
        float n5 = 0.0f;
        CoreMetrics coreMetrics2 = null;
        boolean n6 = false;
        this.locs = new float[this.fComponents.length * 2 + 2];
        TextLineComponent textLineComponent2;
        CoreMetrics coreMetrics3;
        float effectiveBaselineOffset;
        for (int j = 0, n7 = 0; j < this.fComponents.length; this.locs[n7] = n5, this.locs[n7 + 1] = effectiveBaselineOffset, n5 += textLineComponent2.getAdvance(), coreMetrics2 = coreMetrics3, n6 |= (textLineComponent2.getBaselineTransform() != null), ++j, n7 += 2) {
            textLineComponent2 = this.fComponents[this.getComponentLogicalIndex(j)];
            coreMetrics3 = textLineComponent2.getCoreMetrics();
            if (coreMetrics2 != null && (coreMetrics2.italicAngle != 0.0f || coreMetrics3.italicAngle != 0.0f) && (coreMetrics2.italicAngle != coreMetrics3.italicAngle || coreMetrics2.baselineIndex != coreMetrics3.baselineIndex || coreMetrics2.ssOffset != coreMetrics3.ssOffset)) {
                float effectiveBaselineOffset2 = coreMetrics2.effectiveBaselineOffset(this.fBaselineOffsets);
                float a = effectiveBaselineOffset2 - coreMetrics2.ascent;
                float a2 = effectiveBaselineOffset2 + coreMetrics2.descent;
                float effectiveBaselineOffset3 = coreMetrics3.effectiveBaselineOffset(this.fBaselineOffsets);
                float b4 = effectiveBaselineOffset3 - coreMetrics3.ascent;
                float b5 = effectiveBaselineOffset3 + coreMetrics3.descent;
                float max6 = Math.max(a, b4);
                float min = Math.min(a2, b5);
                n5 += Math.max(coreMetrics2.italicAngle * (effectiveBaselineOffset2 - max6) - coreMetrics3.italicAngle * (effectiveBaselineOffset3 - max6), coreMetrics2.italicAngle * (effectiveBaselineOffset2 - min) - coreMetrics3.italicAngle * (effectiveBaselineOffset3 - min));
                effectiveBaselineOffset = effectiveBaselineOffset3;
            }
            else {
                effectiveBaselineOffset = coreMetrics3.effectiveBaselineOffset(this.fBaselineOffsets);
            }
        }
        if (coreMetrics2.italicAngle != 0.0f) {
            float effectiveBaselineOffset4 = coreMetrics2.effectiveBaselineOffset(this.fBaselineOffsets);
            float n8 = effectiveBaselineOffset4 - coreMetrics2.ascent;
            float n9 = effectiveBaselineOffset4 + coreMetrics2.descent;
            float n10 = effectiveBaselineOffset4 + coreMetrics2.ssOffset;
            float n11;
            if (coreMetrics2.italicAngle > 0.0f) {
                n11 = n10 + coreMetrics2.ascent;
            }
            else {
                n11 = n10 - coreMetrics2.descent;
            }
            n5 += n11 * coreMetrics2.italicAngle;
        }
        this.locs[this.locs.length - 2] = n5;
        this.fMetrics = new TextLineMetrics(max, max2, n4, n5);
        if (n6 != 0) {
            this.isSimple = false;
            Point2D.Double o = new Point2D.Double();
            double n12 = 0.0;
            double n13 = 0.0;
            LayoutPathImpl.SegmentPathBuilder o2 = new LayoutPathImpl.SegmentPathBuilder();
            o2.moveTo((double)this.locs[0], 0.0);
            for (int k = 0, n14 = 0; k < this.fComponents.length; ++k, n14 += 2) {
                AffineTransform baselineTransform = this.fComponents[this.getComponentLogicalIndex(k)].getBaselineTransform();
                if (baselineTransform != null && (baselineTransform.getType() & 0x1) != 0x0) {
                    o2.moveTo(n12 += baselineTransform.getTranslateX(), n13 += baselineTransform.getTranslateY());
                }
                o.x = (double)(this.locs[n14 + 2] - this.locs[n14]);
                o.y = 0.0;
                if (baselineTransform != null) {
                    baselineTransform.deltaTransform(o, o);
                }
                o2.lineTo(n12 += o.x, n13 += o.y);
            }
            this.lp = o2.complete();
            if (this.lp == null) {
                AffineTransform baselineTransform2 = this.fComponents[this.getComponentLogicalIndex(0)].getBaselineTransform();
                if (baselineTransform2 != null) {
                    this.lp = new LayoutPathImpl.EmptyPath(baselineTransform2);
                }
            }
        }
    }
    
    public Rectangle getPixelBounds(FontRenderContext fontRenderContext, float n, float n2) {
        Object computePixelBounds = null;
        if (fontRenderContext != null && fontRenderContext.equals(this.frc)) {
            fontRenderContext = null;
        }
        int n3 = (int)Math.floor((double)n);
        int n4 = (int)Math.floor((double)n2);
        float n5 = n - (float)n3;
        float n6 = n2 - (float)n4;
        int n7 = fontRenderContext == null && n5 == 0.0f && n6 == 0.0f;
        if (n7 && this.pixelBounds != null) {
            Rectangle o2 = new Rectangle(this.pixelBounds);
            o2.x += n3;
            o2.y += n4;
            return o2;
        }
        if (this.isSimple) {
            for (int i = 0, n8 = 0; i < this.fComponents.length; ++i, n8 += 2) {
                Rectangle pixelBounds = this.fComponents[this.getComponentLogicalIndex(i)].getPixelBounds(fontRenderContext, this.locs[n8] + n5, this.locs[n8 + 1] + n6);
                if (!pixelBounds.isEmpty()) {
                    if (computePixelBounds == null) {
                        computePixelBounds = pixelBounds;
                    }
                    else {
                        computePixelBounds.add(pixelBounds);
                    }
                }
            }
            if (computePixelBounds == null) {
                computePixelBounds = new Rectangle(0, 0, 0, 0);
            }
        }
        else {
            Rectangle2D rectangle2D = this.getVisualBounds();
            if (this.lp != null) {
                rectangle2D = this.lp.mapShape(rectangle2D).getBounds();
            }
            Rectangle bounds = rectangle2D.getBounds();
            BufferedImage o = new BufferedImage(bounds.width + 6, bounds.height + 6, 2);
            Graphics2D graphics = o.createGraphics();
            graphics.setColor(Color.WHITE);
            graphics.fillRect(0, 0, o.getWidth(), o.getHeight());
            graphics.setColor(Color.BLACK);
            this.draw(graphics, n5 + 3.0f - (float)bounds.x, n6 + 3.0f - (float)bounds.y);
            computePixelBounds = computePixelBounds(o);
            computePixelBounds.x -= 3 - bounds.x;
            computePixelBounds.y -= 3 - bounds.y;
        }
        if (!(n7 == 0)) {
            this.pixelBounds = new Rectangle(computePixelBounds);
        }
        computePixelBounds.x += n3;
        computePixelBounds.y += n4;
        return computePixelBounds;
    }
    
    static Rectangle computePixelBounds(BufferedImage bufferedImage) {
        int width = bufferedImage.getWidth();
        int height = bufferedImage.getHeight();
        int n = -1;
        int i = -1;
        int[] rgbArray = new int[width];
    Label_0083:
        while (++i < height) {
            bufferedImage.getRGB(0, i, rgbArray.length, 1, rgbArray, 0, width);
            int j = 0;
            while (j < rgbArray.length) {
                if (rgbArray[j] != -1) {
                    break Label_0083;
                }
                ++j;
            }
        }
        int[] rgbArray2 = new int[width];
    Label_0146:
        while (--height > i) {
            bufferedImage.getRGB(0, height, rgbArray2.length, 1, rgbArray2, 0, width);
            int k = 0;
            while (k < rgbArray2.length) {
                if (rgbArray2[k] != -1) {
                    break Label_0146;
                }
                ++k;
            }
        }
        ++height;
    Label_0196:
        while (++n < width) {
            while (i < height) {
                if (bufferedImage.getRGB(n, i) != -1) {
                    break Label_0196;
                }
                ++i;
            }
        }
    Label_0244:
        while (--width > n) {
            while (i < height) {
                if (bufferedImage.getRGB(width, i) != -1) {
                    break Label_0244;
                }
                ++i;
            }
        }
        ++width;
        return new Rectangle(n, i, width - n, height - i);
    }
    
    public int characterCount() {
        return this.fCharsLimit - this.fCharsStart;
    }
    
    public boolean isDirectionLTR() {
        return this.fIsDirectionLTR;
    }
    
    public TextLine.TextLineMetrics getMetrics() {
        return this.fMetrics;
    }
    
    public int visualToLogical(int n) {
        if (this.fCharLogicalOrder == null) {
            return n;
        }
        if (this.fCharVisualOrder == null) {
            this.fCharVisualOrder = BidiUtils.createInverseMap(this.fCharLogicalOrder);
        }
        return this.fCharVisualOrder[n];
    }
    
    public int logicalToVisual(int n) {
        return (this.fCharLogicalOrder == null) ? n : this.fCharLogicalOrder[n];
    }
    
    public byte getCharLevel(int n) {
        return (this.fCharLevels == null) ? 0 : this.fCharLevels[n];
    }
    
    public boolean isCharLTR(int n) {
        return (this.getCharLevel(n) & 0x1) == 0x0;
    }
    
    public int getCharType(int n) {
        return Character.getType(this.fChars[n + this.fCharsStart]);
    }
    
    public boolean isCharSpace(int n) {
        return Character.isSpaceChar(this.fChars[n + this.fCharsStart]);
    }
    
    public boolean isCharWhitespace(int n) {
        return Character.isWhitespace(this.fChars[n + this.fCharsStart]);
    }
    
    public float getCharAngle(int n) {
        return this.getCoreMetricsAt(n).italicAngle;
    }
    
    public CoreMetrics getCoreMetricsAt(int n) {
        if (n < 0) {
            throw new IllegalArgumentException("Negative logicalIndex.");
        }
        if (n > this.fCharsLimit - this.fCharsStart) {
            throw new IllegalArgumentException("logicalIndex too large.");
        }
        int i = 0;
        int n2 = 0;
        do {
            n2 += this.fComponents[i].getNumCharacters();
            if (n2 > n) {
                break;
            }
            ++i;
        } while (i < this.fComponents.length);
        return this.fComponents[i].getCoreMetrics();
    }
    
    public float getCharAscent(int n) {
        return this.getCoreMetricsAt(n).ascent;
    }
    
    public float getCharDescent(int n) {
        return this.getCoreMetricsAt(n).descent;
    }
    
    public float getCharShift(int n) {
        return this.getCoreMetricsAt(n).ssOffset;
    }
    
    private float applyFunctionAtIndex(int n, TextLine.Function function) {
        if (n < 0) {
            throw new IllegalArgumentException("Negative logicalIndex.");
        }
        int n2 = 0;
        int i = 0;
        while (i < this.fComponents.length) {
            int n3 = n2 + this.fComponents[i].getNumCharacters();
            if (n3 > n) {
                return function.computeFunction(this, i, n - n2);
            }
            n2 = n3;
            ++i;
        }
        throw new IllegalArgumentException("logicalIndex too large.");
    }
    
    public float getCharAdvance(int n) {
        return this.applyFunctionAtIndex(n, TextLine.fgAdvanceF);
    }
    
    public float getCharXPosition(int n) {
        return this.applyFunctionAtIndex(n, TextLine.fgXPositionF);
    }
    
    public float getCharYPosition(int n) {
        return this.applyFunctionAtIndex(n, TextLine.fgYPositionF);
    }
    
    public float getCharLinePosition(int n) {
        return this.getCharXPosition(n);
    }
    
    public float getCharLinePosition(int n, boolean b) {
        return this.applyFunctionAtIndex(n, (this.isCharLTR(n) == b) ? TextLine.fgXPositionF : TextLine.fgPosAdvF);
    }
    
    public boolean caretAtOffsetIsValid(int n) {
        if (n < 0) {
            throw new IllegalArgumentException("Negative offset.");
        }
        int n2 = 0;
        int i = 0;
        while (i < this.fComponents.length) {
            int n3 = n2 + this.fComponents[i].getNumCharacters();
            if (n3 > n) {
                return this.fComponents[i].caretAtOffsetIsValid(n - n2);
            }
            n2 = n3;
            ++i;
        }
        throw new IllegalArgumentException("logicalIndex too large.");
    }
    
    private int getComponentLogicalIndex(int n) {
        if (this.fComponentVisualOrder == null) {
            return n;
        }
        return this.fComponentVisualOrder[n];
    }
    
    private int getComponentVisualIndex(int i) {
        if (this.fComponentVisualOrder == null) {
            return i;
        }
        int j = 0;
        while (j < this.fComponentVisualOrder.length) {
            if (this.fComponentVisualOrder[j] == i) {
                return j;
            }
            ++j;
        }
        throw new IndexOutOfBoundsException("bad component index: " + i);
    }
    
    public Rectangle2D getCharBounds(int n) {
        if (n < 0) {
            throw new IllegalArgumentException("Negative logicalIndex.");
        }
        int n2 = 0;
        int i = 0;
        while (i < this.fComponents.length) {
            int n3 = n2 + this.fComponents[i].getNumCharacters();
            if (n3 > n) {
                Rectangle2D charVisualBounds = this.fComponents[i].getCharVisualBounds(n - n2);
                int componentVisualIndex = this.getComponentVisualIndex(i);
                charVisualBounds.setRect(charVisualBounds.getX() + (double)this.locs[componentVisualIndex * 2], charVisualBounds.getY() + (double)this.locs[componentVisualIndex * 2 + 1], charVisualBounds.getWidth(), charVisualBounds.getHeight());
                return charVisualBounds;
            }
            n2 = n3;
            ++i;
        }
        throw new IllegalArgumentException("logicalIndex too large.");
    }
    
    private float getComponentShift(int n) {
        return this.fComponents[n].getCoreMetrics().effectiveBaselineOffset(this.fBaselineOffsets);
    }
    
    public void draw(Graphics2D graphics2D, float n, float n2) {
        if (this.lp == null) {
            for (int j = 0, n4 = 0; j < this.fComponents.length; ++j, n4 += 2) {
                this.fComponents[this.getComponentLogicalIndex(j)].draw(graphics2D, this.locs[n4] + n, this.locs[n4 + 1] + n2);
            }
        }
        else {
            AffineTransform transform = graphics2D.getTransform();
            Point2D.Float pt = new Point2D.Float();
            for (int i = 0, n3 = 0; i < this.fComponents.length; ++i, n3 += 2) {
                TextLineComponent textLineComponent = this.fComponents[this.getComponentLogicalIndex(i)];
                this.lp.pathToPoint((double)this.locs[n3], (double)this.locs[n3 + 1], false, pt);
                pt.x += n;
                pt.y += n2;
                AffineTransform baselineTransform = textLineComponent.getBaselineTransform();
                if (baselineTransform != null) {
                    graphics2D.translate((double)pt.x - baselineTransform.getTranslateX(), (double)pt.y - baselineTransform.getTranslateY());
                    graphics2D.transform(baselineTransform);
                    textLineComponent.draw(graphics2D, 0.0f, 0.0f);
                    graphics2D.setTransform(transform);
                }
                else {
                    textLineComponent.draw(graphics2D, pt.x, pt.y);
                }
            }
        }
    }
    
    public Rectangle2D getVisualBounds() {
        Object o = null;
        for (int i = 0, n = 0; i < this.fComponents.length; ++i, n += 2) {
            TextLineComponent textLineComponent = this.fComponents[this.getComponentLogicalIndex(i)];
            Rectangle2D rectangle2D = textLineComponent.getVisualBounds();
            Point2D.Float o2 = new Point2D.Float(this.locs[n], this.locs[n + 1]);
            if (this.lp == null) {
                rectangle2D.setRect(rectangle2D.getMinX() + (double)o2.x, rectangle2D.getMinY() + (double)o2.y, rectangle2D.getWidth(), rectangle2D.getHeight());
            }
            else {
                this.lp.pathToPoint(o2, false, o2);
                AffineTransform baselineTransform = textLineComponent.getBaselineTransform();
                if (baselineTransform != null) {
                    AffineTransform translateInstance = AffineTransform.getTranslateInstance((double)o2.x - baselineTransform.getTranslateX(), (double)o2.y - baselineTransform.getTranslateY());
                    translateInstance.concatenate(baselineTransform);
                    rectangle2D = translateInstance.createTransformedShape(rectangle2D).getBounds2D();
                }
                else {
                    rectangle2D.setRect(rectangle2D.getMinX() + (double)o2.x, rectangle2D.getMinY() + (double)o2.y, rectangle2D.getWidth(), rectangle2D.getHeight());
                }
            }
            if (o == null) {
                o = rectangle2D;
            }
            else {
                o.add(rectangle2D);
            }
        }
        if (o == null) {
            o = new Rectangle2D.Float(Float.MAX_VALUE, Float.MAX_VALUE, Float.MIN_VALUE, Float.MIN_VALUE);
            return o;
        }
        return o;
    }
    
    public Rectangle2D getItalicBounds() {
        float min = Float.MAX_VALUE;
        float max = -3.4028235E38f;
        float min2 = Float.MAX_VALUE;
        float max2 = -3.4028235E38f;
        for (int i = 0, n = 0; i < this.fComponents.length; ++i, n += 2) {
            Rectangle2D italicBounds = this.fComponents[this.getComponentLogicalIndex(i)].getItalicBounds();
            float n2 = this.locs[n];
            float n3 = this.locs[n + 1];
            min = Math.min(min, n2 + (float)italicBounds.getX());
            max = Math.max(max, n2 + (float)italicBounds.getMaxX());
            min2 = Math.min(min2, n3 + (float)italicBounds.getY());
            max2 = Math.max(max2, n3 + (float)italicBounds.getMaxY());
        }
        return new Rectangle2D.Float(min, min2, max - min, max2 - min2);
    }
    
    public Shape getOutline(AffineTransform at) {
        GeneralPath generalPath = new GeneralPath(1);
        for (int i = 0, n = 0; i < this.fComponents.length; ++i, n += 2) {
            generalPath.append(this.fComponents[this.getComponentLogicalIndex(i)].getOutline(this.locs[n], this.locs[n + 1]), false);
        }
        if (at != null) {
            generalPath.transform(at);
            return generalPath;
        }
        return generalPath;
    }
    
    @Override
    public int hashCode() {
        return this.fComponents.length << 16 ^ this.fComponents[0].hashCode() << 3 ^ this.fCharsLimit - this.fCharsStart;
    }
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < this.fComponents.length; ++i) {
            sb.append(this.fComponents[i]);
        }
        return sb.toString();
    }
    
    public static TextLine fastCreateTextLine(FontRenderContext frc, char[] text, Font font, CoreMetrics coreMetrics, Map<? extends AttributedCharacterIterator.Attribute, ?> map) {
        boolean baseIsLeftToRight = true;
        byte[] levels = null;
        int[] inverseMap = null;
        Object o = null;
        int length = text.length;
        boolean requiresBidi = false;
        byte[] embeddings = null;
        AttributeValues fromMap = null;
        if (map != null) {
            fromMap = AttributeValues.fromMap(map);
            if (fromMap.getRunDirection() >= 0) {
                baseIsLeftToRight = (fromMap.getRunDirection() == 0);
                requiresBidi = (baseIsLeftToRight ? false : true);
            }
            if (fromMap.getBidiEmbedding() != 0) {
                requiresBidi = true;
                int n = (byte)fromMap.getBidiEmbedding();
                embeddings = new byte[length];
                for (int i = 0; i < embeddings.length; ++i) {
                    embeddings[i] = n;
                }
            }
        }
        if (requiresBidi == 0) {
            requiresBidi = Bidi.requiresBidi(text, 0, text.length);
        }
        if (requiresBidi != 0) {
            o = new Bidi(text, 0, embeddings, 0, text.length, (fromMap == null) ? -2 : fromMap.getRunDirection());
            if (!o.isLeftToRight()) {
                levels = BidiUtils.getLevels(o);
                inverseMap = BidiUtils.createInverseMap(BidiUtils.createVisualToLogicalMap(levels));
                baseIsLeftToRight = o.baseIsLeftToRight();
            }
        }
        TextLineComponent[] componentsOnRun;
        int length2;
        for (componentsOnRun = createComponentsOnRun(0, text.length, text, inverseMap, levels, new TextLabelFactory(frc, text, o, 0), font, coreMetrics, frc, Decoration.getDecoration(fromMap), new TextLineComponent[1], 0), length2 = componentsOnRun.length; componentsOnRun[length2 - 1] == null; --length2) {}
        if (length2 != componentsOnRun.length) {
            TextLineComponent[] array = new TextLineComponent[length2];
            System.arraycopy(componentsOnRun, 0, array, 0, length2);
            componentsOnRun = array;
        }
        return new TextLine(frc, componentsOnRun, coreMetrics.baselineOffsets, text, 0, text.length, inverseMap, levels, (boolean)baseIsLeftToRight);
    }
    
    private static TextLineComponent[] expandArray(TextLineComponent[] array) {
        TextLineComponent[] array2 = new TextLineComponent[array.length + 8];
        System.arraycopy(array, 0, array2, 0, array.length);
        return array2;
    }
    
    public static TextLineComponent[] createComponentsOnRun(int i, int n, char[] chars, int[] array, byte[] array2, TextLabelFactory textLabelFactory, Font font, CoreMetrics value, FontRenderContext frc, Decoration decorator, TextLineComponent[] expandArray, int n2) {
        do {
            int firstVisualChunk = firstVisualChunk(array, array2, i, n);
            do {
                int numChars;
                if (value == null) {
                    LineMetrics lineMetrics = font.getLineMetrics(chars, i, firstVisualChunk, frc);
                    value = CoreMetrics.get(lineMetrics);
                    numChars = lineMetrics.getNumChars();
                }
                else {
                    numChars = firstVisualChunk - i;
                }
                ExtendedTextLabel extended = textLabelFactory.createExtended(font, value, decorator, i, i + numChars);
                if (++n2 >= expandArray.length) {
                    expandArray = expandArray(expandArray);
                }
                expandArray[n2 - 1] = extended;
                i += numChars;
            } while (i < firstVisualChunk);
        } while (i < n);
        return expandArray;
    }
    
    public static TextLineComponent[] getComponents(StyledParagraph styledParagraph, char[] array, int i, int b, int[] charsLtoV, byte[] levels, TextLabelFactory textLabelFactory) {
        FontRenderContext fontRenderContext = textLabelFactory.getFontRenderContext();
        int length = 0;
        TextLineComponent[] array2 = { null };
        do {
            int min = Math.min(styledParagraph.getRunLimit(i), b);
            Decoration decoration = styledParagraph.getDecorationAt(i);
            Object fontOrGraphic = styledParagraph.getFontOrGraphicAt(i);
            if ((fontOrGraphic instanceof GraphicAttribute)) {
                __Null baseTx = null;
                GraphicAttribute graphic = (GraphicAttribute)fontOrGraphic;
                do {
                    int firstVisualChunk = firstVisualChunk(charsLtoV, levels, i, min);
                    GraphicComponent o = new GraphicComponent(graphic, decoration, charsLtoV, levels, i, firstVisualChunk, baseTx);
                    i = firstVisualChunk;
                    if (++length >= array2.length) {
                        array2 = expandArray(array2);
                    }
                    array2[length - 1] = o;
                } while (i < min);
            }
            else {
                array2 = createComponentsOnRun(i, min, array, charsLtoV, levels, textLabelFactory, (Font)fontOrGraphic, null, fontRenderContext, decoration, array2, length);
                i = min;
                for (length = array2.length; array2[length - 1] == null; --length) {}
            }
        } while (i < b);
        if (array2.length == length) {
            return array2;
        }
        array2 = new TextLineComponent[length];
        System.arraycopy(array2, 0, array2, 0, length);
        return array2;
    }
    
    public static TextLine createLineFromText(char[] array, StyledParagraph styledParagraph, TextLabelFactory textLabelFactory, boolean b, float[] array2) {
        textLabelFactory.setLineContext(0, array.length);
        Bidi lineBidi = textLabelFactory.getLineBidi();
        int[] inverseMap = null;
        byte[] levels = null;
        if (lineBidi != null) {
            levels = BidiUtils.getLevels(lineBidi);
            inverseMap = BidiUtils.createInverseMap(BidiUtils.createVisualToLogicalMap(levels));
        }
        return new TextLine(textLabelFactory.getFontRenderContext(), getComponents(styledParagraph, array, 0, array.length, inverseMap, levels, textLabelFactory), array2, array, 0, array.length, inverseMap, levels, b);
    }
    
    private static int[] computeComponentOrder(TextLineComponent[] array, int[] array2) {
        int[] inverseMap = null;
        if (array2 != null && array.length > 1) {
            int[] values = new int[array.length];
            int n = 0;
            for (int i = 0; i < array.length; ++i) {
                values[i] = array2[n];
                n += array[i].getNumCharacters();
            }
            inverseMap = BidiUtils.createInverseMap(BidiUtils.createContiguousOrder(values));
            return inverseMap;
        }
        return inverseMap;
    }
    
    public static TextLine standardCreateTextLine(FontRenderContext frc, AttributedCharacterIterator paragraph, char[] text, float[] array) {
        StyledParagraph styledParagraph = new StyledParagraph(paragraph, text);
        Bidi bidi = new Bidi(paragraph);
        if (bidi.isLeftToRight()) {
            bidi = null;
        }
        TextLabelFactory o = new TextLabelFactory(frc, text, bidi, 0);
        boolean baseIsLeftToRight = true;
        if (bidi != null) {
            baseIsLeftToRight = bidi.baseIsLeftToRight();
        }
        return createLineFromText(text, styledParagraph, o, (boolean)baseIsLeftToRight, array);
    }
    
    static boolean advanceToFirstFont(AttributedCharacterIterator attributedCharacterIterator) {
        int i = attributedCharacterIterator.first();
        while (i != '\uffff') {
            if (attributedCharacterIterator.getAttribute(TextAttribute.CHAR_REPLACEMENT) == null) {
                return true;
            }
            i = attributedCharacterIterator.setIndex(attributedCharacterIterator.getRunLimit());
        }
        return false;
    }
    
    static float[] getNormalizedOffsets(float[] array, byte b) {
        if (array[b] != 0.0f) {
            float n = array[b];
            float[] array2 = new float[array.length];
            for (int i = 0; i < array2.length; ++i) {
                array2[i] = array[i] - n;
            }
            array = array2;
            return array;
        }
        return array;
    }
    
    static Font getFontAtCurrentPos(AttributedCharacterIterator attributedCharacterIterator) {
        Object attribute = attributedCharacterIterator.getAttribute(TextAttribute.FONT);
        if (attribute != null) {
            return (Font)attribute;
        }
        if (attributedCharacterIterator.getAttribute(TextAttribute.FAMILY) != null) {
            return Font.getFont(attributedCharacterIterator.getAttributes());
        }
        int next = CodePointIterator.create(attributedCharacterIterator).next();
        if (next != -1) {
            FontResolver instance = FontResolver.getInstance();
            return instance.getFont(instance.getFontIndex(next), attributedCharacterIterator.getAttributes());
        }
        return null;
    }
    
    private static int firstVisualChunk(int[] array, byte[] array2, int n, int n2) {
        if (array != null && array2 != null) {
            int n3 = array2[n];
            while (++n < n2 && array2[n] == n3) {}
            return n;
        }
        return n2;
    }
    
    public TextLine getJustifiedLine(float n, float n2, int n3, int n4) {
        TextLineComponent[] array = new TextLineComponent[this.fComponents.length];
        System.arraycopy(this.fComponents, 0, array, 0, this.fComponents.length);
        int i = 0;
        do {
            getAdvanceBetween(array, 0, this.characterCount());
            float n5 = (n - getAdvanceBetween(array, n3, n4)) * n2;
            int[] array2 = new int[array.length];
            int n6 = 0;
            for (int j = 0; j < array.length; ++j) {
                int componentLogicalIndex = this.getComponentLogicalIndex(j);
                array2[componentLogicalIndex] = n6;
                n6 += array[componentLogicalIndex].getNumJustificationInfos();
            }
            GlyphJustificationInfo[] array3 = new GlyphJustificationInfo[n6];
            int n7 = 0;
            for (int k = 0; k < array.length; ++k) {
                TextLineComponent textLineComponent = array[k];
                int numCharacters = textLineComponent.getNumCharacters();
                int n8 = n7 + numCharacters;
                if (n8 > n3) {
                    textLineComponent.getJustificationInfos(array3, array2[k], Math.max(0, n3 - n7), Math.min(numCharacters, n4 - n7));
                    if (n8 >= n4) {
                        break;
                    }
                }
            }
            int n9;
            for (n9 = 0, ; n9 < n6 && array3[n9] == null; ++n9) {}
            while (n6 > n9 && array3[n6 - 1] == null) {
                --n6;
            }
            float[] justify = new TextJustifier(array3, n9, n6).justify(n5);
            int n10 = i == 0;
            boolean n11 = false;
            boolean[] array4 = { false };
            int n12 = 0;
            for (int l = 0; l < array.length; ++l) {
                TextLineComponent textLineComponent2 = array[l];
                int numCharacters2 = textLineComponent2.getNumCharacters();
                int n13 = n12 + numCharacters2;
                if (n13 > n3) {
                    Math.max(0, n3 - n12);
                    Math.min(numCharacters2, n4 - n12);
                    array[l] = textLineComponent2.applyJustificationDeltas(justify, array2[l] * 2, array4);
                    n11 |= array4[0];
                    if (n13 >= n4) {
                        break;
                    }
                }
            }
            i = (n11 && i == 0);
        } while (i != 0);
        return new TextLine(this.frc, array, this.fBaselineOffsets, this.fChars, this.fCharsStart, this.fCharsLimit, this.fCharLogicalOrder, this.fCharLevels, this.fIsDirectionLTR);
    }
    
    public static float getAdvanceBetween(TextLineComponent[] array, int n, int n2) {
        float n3 = 0.0f;
        int n4 = 0;
        int n5;
        for (int i = 0; i < array.length; n4 = n5, ++i) {
            TextLineComponent textLineComponent = array[i];
            int numCharacters = textLineComponent.getNumCharacters();
            n5 = n4 + numCharacters;
            if (n5 > n) {
                n3 += textLineComponent.getAdvanceBetween(Math.max(0, n - n4), Math.min(numCharacters, n2 - n4));
                if (n5 >= n2) {
                    break;
                }
            }
        }
        return n3;
    }
    
    LayoutPathImpl getLayoutPath() {
        return this.lp;
    }
    
    static int access$200(TextLine textLine, int n) {
        return textLine.getComponentVisualIndex(n);
    }
    
    static float access$400(TextLine textLine, int n) {
        return textLine.getComponentShift(n);
    }
    
    static {
        TextLine.fgPosAdvF = new TextLine.Function() {
            @Override
            float computeFunction(TextLine textLine, int n, int n2) {
                TextLineComponent textLineComponent = textLine.fComponents[n];
                return textLine.locs[TextLine.access$200(textLine, n) * 2] + textLineComponent.getCharX(n2) + textLineComponent.getCharAdvance(n2);
            }
        };
        TextLine.fgAdvanceF = new TextLine.Function() {
            @Override
            float computeFunction(TextLine textLine, int n, int n2) {
                return textLine.fComponents[n].getCharAdvance(n2);
            }
        };
        TextLine.fgXPositionF = new TextLine.Function() {
            @Override
            float computeFunction(TextLine textLine, int n, int n2) {
                return textLine.locs[TextLine.access$200(textLine, n) * 2] + textLine.fComponents[n].getCharX(n2);
            }
        };
        TextLine.fgYPositionF = new TextLine.Function() {
            @Override
            float computeFunction(TextLine textLine, int n, int n2) {
                return textLine.fComponents[n].getCharY(n2) + TextLine.access$400(textLine, n);
            }
        };
    }
    
    static final class TextLineMetrics
    {
        public final float ascent;
        public final float descent;
        public final float leading;
        public final float advance;
        
        public TextLineMetrics(float ascent, float descent, float leading, float advance) {
            this.ascent = ascent;
            this.descent = descent;
            this.leading = leading;
            this.advance = advance;
        }
    }
    
    private abstract static class Function
    {
        abstract float computeFunction(TextLine p0, int p1, int p2);
    }
}
